import json
import logging
import os
import subprocess
from typing import Optional


class MetadataWriter:
    SUPPORTED_EXTENSIONS = (".mp4", ".mkv", ".mov", ".avi")

    def __init__(self):
        self.logger: logging.Logger = logging.getLogger("MetadataWriter")

    @staticmethod
    def get_mkvmerge_cmd(metadata: dict, in_path: str, out_path: str) -> list[str]:
        cmd: list[str] = [
            "mkvmerge",
            "-o", out_path
        ]

        # Add global metadata (title)
        if "title" in metadata:
            cmd.extend(["--title", metadata["title"]])

        # Process audio + subtitle tracks
        tracks: list[dict] = metadata.get("audio_tracks", []) + metadata.get("subtitle_tracks", [])
        for track in tracks:
            # Use the actual track index from the metadata
            track_id = track.get("index", 0)

            # Set language
            if "language" in track:
                cmd.extend(["--language", f"{track_id}:{track["language"]}"])

            # Set title/name
            if "name" in track and track["name"]:
                cmd.extend(["--track-name", f"{track_id}:{track["name"]}"])

            # Set disposition flags
            flags = track.get("flags", {})

            def yes_no(flag: str):
                return f"{track_id}:{"yes" if flags.get(flag, False) else "no"}"

            cmd.extend(["--default-track", yes_no("default")])
            cmd.extend(["--forced-track", yes_no("forced")])
            cmd.extend(["--original-flag", yes_no("original")])

        # Add input file
        cmd.append(in_path)
        return cmd

    @staticmethod
    def get_mkvpropedit_cmd(metadata: dict, path: str) -> list[str]:
        cmd: list[str] = [
            "mkvpropedit",
            path
        ]

        # Add global metadata (title)
        if "title" in metadata:
            cmd.extend(["--edit", "info", "--set", f"title={metadata["title"]}"])

        # Process audio + subtitle tracks
        tracks: list[dict] = metadata.get("audio_tracks", []) + metadata.get("subtitle_tracks", [])
        for track in tracks:
            # Use the actual track index from the metadata
            track_id = track.get("index", 0)

            cmd.extend(["--edit", f"track:{track_id}"])

            # Set language
            if "language" in track:
                cmd.extend(["--set", f"language={track["language"]}"])

            # Set title/name
            if "name" in track and track["name"]:
                cmd.extend(["--set", f"name={track["name"]}"])

            # Set disposition flags
            flags = track.get("flags", {})

            cmd.extend(["--set", f"flag-default={int(flags.get("default", False))}"])
            cmd.extend(["--set", f"flag-forced={int(flags.get("forced", False))}"])
            cmd.extend(["--set", f"flag-original={int(flags.get("original", False))}"])

        return cmd

    def apply_metadata(self, metadata: dict, in_path: str, out_path: Optional[str] = None) -> bool:
        """
        Writes metadata to a video file using mkvmerge or mkvpropedit

        :param metadata: Metadata information
        :param in_path: Path of the input video file
        :param out_path: Path of the output video file. If None, ``"_modified"`` is appended to ``in_path`` instead
        :return: True if successful, False otherwise
        """

        if not os.path.isfile(in_path):
            self.logger.error(f"Input file not found: {in_path}")
            return False

        if out_path is None:
            # Create a temporary output file
            base_name, ext = os.path.splitext(in_path)
            out_path: str = f"{base_name}_modified{ext}"

        # Build the command
        overwriting: bool = os.path.abspath(in_path) == os.path.abspath(out_path)
        cmd: list[str] = (
            self.get_mkvpropedit_cmd(metadata, in_path)
            if overwriting else
            self.get_mkvmerge_cmd(metadata, in_path, out_path)
        )

        # Execute the command
        self.logger.debug(f"Writing metadata to {os.path.basename(out_path)}")

        try:
            result = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
            if result.returncode != 0:
                self.logger.error(f"Error writing metadata: {result.stderr}")
                return False

            self.logger.debug(f"Metadata written to {out_path}")
            return True

        except Exception as e:
            self.logger.error(f"Error executing {cmd[0]}: {str(e)}")
            return False

    @staticmethod
    def read_metadata(path: str) -> Optional[dict]:
        try:
            with open(path, "r") as f:
                metadata: dict = json.load(f)
            return metadata

        except:
            return None

    def process_file(self, metadata_or_path: str|dict, file_path: str, output_dir: Optional[str] = None) -> bool:
        """
        Processes a single video file with the given metadata

        :param metadata_or_path: Metadata dict or path of the metadata file
        :param file_path: Path of the video file
        :param output_dir: Directory to save the output file to
        :return: True if successful, False otherwise
        """

        metadata: dict
        if isinstance(metadata_or_path, str):
            metadata = self.read_metadata(metadata_or_path)
            if metadata is None:
                return False
        else:
            metadata = metadata_or_path

        # Create output file path
        if output_dir is not None:
            # Ensure output directory exists
            os.makedirs(output_dir, exist_ok=True)

            # Use the same filename in the output directory
            output_file = os.path.join(output_dir, os.path.basename(file_path))
        else:
            output_file = None

        # Write metadata to video
        return self.apply_metadata(metadata, file_path, output_file)

    def process_directory(self, metadata_or_path: str|dict, source_dir: str, output_dir: Optional[str] = None) -> bool:
        """
        Processes all video files in the metadata dictionary

        :param metadata_or_path: Dictionary of metadata keyed by filename
        :param source_dir: Directory containing the video files
        :param output_dir: Directory to save the output files to
        :return: True if all files were processed successfully, False otherwise
        """

        metadata: dict
        if isinstance(metadata_or_path, str):
            metadata = self.read_metadata(metadata_or_path)
            if metadata is None:
                return False
        else:
            metadata = metadata_or_path

        if not os.path.isdir(source_dir):
            self.logger.error(f"Source directory not found: {source_dir}")
            return False

        # Create output directory if specified
        if output_dir:
            os.makedirs(output_dir, exist_ok=True)

        success: bool = True
        processed_count: int = 0

        # Process each file in the metadata dictionary
        for filename, file_metadata in metadata.items():
            # Construct the full path to the video file
            video_file: str = os.path.join(source_dir, filename)

            if not os.path.isfile(video_file):
                self.logger.error(f"Video file not found: {video_file}")
                success = False
                continue

            # Process the file
            if self.process_file(file_metadata, video_file, output_dir):
                processed_count += 1
            else:
                success = False

        self.logger.debug(f"Processed {processed_count} out of {len(metadata)} files")
        return success

    def process_metadata(self, metadata_or_path: str|dict, source_dir: Optional[str] = None, output_dir: Optional[str] = None) -> bool:
        metadata_as_path: bool = isinstance(metadata_or_path, str)

        metadata: dict
        if metadata_as_path:
            metadata = self.read_metadata(metadata_or_path)
            if metadata is None:
                return False
        else:
            metadata = metadata_or_path

        # Determine if the JSON contains metadata for multiple files or a single file
        is_multi_file = isinstance(metadata, dict) and all(isinstance(metadata[key], dict) for key in metadata)

        # If source directory is not specified, try to determine it from the JSON filename
        if source_dir is None and is_multi_file and metadata_as_path:
            # Extract folder name from JSON filename (e.g., "Millenium" from "Millenium_metadata.json")
            json_basename: str = os.path.basename(metadata_or_path)
            if json_basename.endswith("_metadata.json"):
                folder_name: str = json_basename.split("_metadata.json")[0]
                potential_source_dir: str = os.path.join(
                    os.path.dirname(os.path.abspath(metadata_or_path)),
                    folder_name
                )

                if os.path.isdir(potential_source_dir):
                    source_dir: str = potential_source_dir
                    self.logger.debug(f"Using source directory: {source_dir}")

        # If no output directory is specified, create one based on the source directory
        if output_dir is None and source_dir is not None:
            output_dir = os.path.join("ready", os.path.basename(source_dir))
            self.logger.debug(f"Using output directory: {output_dir}")

        # Process files based on the metadata format
        if is_multi_file:
            if source_dir is None:
                self.logger.error(
                    "Source directory not specified and could not be determined automatically. " +
                    "Please specify a source directory with --source or use a JSON filename like 'FolderName_metadata.json'"
                )
                return False

            success = self.process_directory(metadata, source_dir, output_dir)
        else:
            # Single file metadata
            if "filename" not in metadata:
                self.logger.error("Invalid metadata format: missing 'filename' field")
                return False

            # If source directory is specified, look for the file there
            video_file: str
            if source_dir is not None:
                video_file = os.path.join(source_dir, metadata["filename"])
            elif metadata_as_path:
                # Look for the file in the same directory as the JSON
                video_file = os.path.join(os.path.dirname(metadata_or_path), metadata["filename"])
            else:
                self.logger.error(
                    "Source directory not specified and video path could not be determined automatically. " +
                    "Please specify a source directory with --source or use JSON filename like 'VideoName_metadata.json'"
                )
                return False

            success = self.process_file(metadata, video_file, output_dir)
        return success